"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.UtxoTransaction = void 0;
const genericMccTypes_1 = require("../../types/genericMccTypes");
const constants_1 = require("../../utils/constants");
const errors_1 = require("../../utils/errors");
const utils_1 = require("../../utils/utils");
const utxoUtils_1 = require("../../utils/utxoUtils");
const TransactionBase_1 = require("../TransactionBase");
// Transaction types and their description
// - coinbase        : transaction that mints new coins
// - payment         : what you get from node
// - partial_payment : transaction with some vout of vins added to additional data
// - full_payment    : transaction with vouts for all vins added to additional data
//@Managed()
class UtxoTransaction extends TransactionBase_1.TransactionBase {
    constructor() {
        super(...arguments);
        this.reducerFunctionAdditionalDataVinOuts = (prev, vout) => { var _a; return prev.add(this.toBnValue((_a = vout === null || vout === void 0 ? void 0 : vout.vinvout) === null || _a === void 0 ? void 0 : _a.value)); };
        this.reducerFunctionPrevouts = (prev, vin) => { var _a; return prev.add(this.toBnValue((_a = vin === null || vin === void 0 ? void 0 : vin.prevout) === null || _a === void 0 ? void 0 : _a.value)); };
        this.reducerFunctionVouts = (prev, vout) => prev.add(this.toBnValue(vout.value));
    }
    get data() {
        return this.privateData;
    }
    get additionalData() {
        return this.privateAdditionalData;
    }
    set additionalData(data) {
        this.privateAdditionalData = data;
    }
    // constructor(data: IUtxoGetTransactionRes, additionalData?: IUtxoTransactionAdditionalData) {
    //     super(data, additionalData);
    //     this.data.vout.forEach((vout) => {
    //         this.processOutput(vout);
    //     });
    //     this.synchronizeAdditionalData();
    //     this.assertAdditionalData();
    // }
    get txid() {
        return this.data.txid;
    }
    get stdTxid() {
        return (0, utils_1.unPrefix0x)(this.txid);
    }
    get hash() {
        return this.data.hash;
    }
    get reference() {
        const references = [];
        for (const vo of this.data.vout) {
            if (vo.scriptPubKey.hex.substring(0, 2) == (0, utils_1.unPrefix0x)((0, utils_1.toHex)(utxoUtils_1.WordToOpcode.OP_RETURN))) {
                const dataSplit = vo.scriptPubKey.asm.split(" ");
                references.push(dataSplit[1]);
            }
        }
        return references;
    }
    get stdPaymentReference() {
        let paymentReference = this.reference.length === 1 ? (0, utils_1.prefix0x)(this.reference[0]) : "";
        if (!(0, utils_1.isValidBytes32Hex)(paymentReference)) {
            paymentReference = utils_1.ZERO_BYTES_32;
        }
        return paymentReference;
    }
    get unixTimestamp() {
        return this.data.blocktime;
    }
    get sourceAddresses() {
        if (this.type === "coinbase") {
            // Coinbase transactions mint coins
            return [undefined];
        }
        if (!this.additionalData || !this.additionalData.vinouts) {
            return [undefined];
        }
        return this.additionalData.vinouts.map((mapper) => { var _a; return (_a = mapper === null || mapper === void 0 ? void 0 : mapper.vinvout) === null || _a === void 0 ? void 0 : _a.scriptPubKey.address; });
    }
    get receivingAddresses() {
        /* istanbul ignore if */
        // Transaction should always have vout array
        if (!this.data.vout) {
            return [];
        }
        return this.data.vout.map((vout) => vout.scriptPubKey.address);
    }
    toBnValue(value) {
        if (value === undefined) {
            return (0, utils_1.toBN)(0);
        }
        return (0, utils_1.toBN)(Math.round(value * constants_1.BTC_MDU).toFixed(0));
    }
    get fee() {
        if (this.type === "full_payment") {
            if (!this.additionalData || !this.additionalData.vinouts) {
                throw new errors_1.mccError(errors_1.mccErrorCode.InvalidTransaction, Error(`Transaction was not made full`));
            }
            const inSum = this.additionalData.vinouts.reduce(this.reducerFunctionAdditionalDataVinOuts, (0, utils_1.toBN)(0));
            const outSum = this.data.vout.reduce(this.reducerFunctionVouts, (0, utils_1.toBN)(0));
            return inSum.sub(outSum);
        }
        if (this.type === "coinbase") {
            // Coinbase transactions mint coins
            return (0, utils_1.toBN)(0);
        }
        throw new errors_1.mccError(errors_1.mccErrorCode.InvalidResponse, Error("fee can't be calculated for `payment` and `partial_payment` transaction types"));
    }
    get feeSignerTotalAmount() {
        throw new Error("Method not implemented.");
    }
    get spentAmounts() {
        if (this.type === "coinbase") {
            // Coinbase transactions mint coins
            return [
                {
                    amount: (0, utils_1.toBN)(0),
                },
            ];
        }
        if (!this.additionalData || !this.additionalData.vinouts) {
            return [
                {
                    amount: (0, utils_1.toBN)(0),
                },
            ];
        }
        return this.additionalData.vinouts.map((mapper) => {
            var _a, _b, _c, _d;
            let amount;
            if (mapper == undefined) {
                amount = (0, utils_1.toBN)(0);
            }
            else {
                amount = (0, utils_1.toBN)(Math.round((((_a = mapper === null || mapper === void 0 ? void 0 : mapper.vinvout) === null || _a === void 0 ? void 0 : _a.value) || 0) * constants_1.BTC_MDU).toFixed(0));
            }
            return {
                address: (_c = (_b = mapper === null || mapper === void 0 ? void 0 : mapper.vinvout) === null || _b === void 0 ? void 0 : _b.scriptPubKey) === null || _c === void 0 ? void 0 : _c.address,
                amount: amount,
                utxo: (_d = mapper === null || mapper === void 0 ? void 0 : mapper.vinvout) === null || _d === void 0 ? void 0 : _d.n,
            };
        });
    }
    get intendedSpentAmounts() {
        return this.spentAmounts;
    }
    get receivedAmounts() {
        return this.data.vout.map((vout) => {
            return {
                address: vout.scriptPubKey.address,
                amount: this.toBnValue(vout.value),
                utxo: vout.n,
            };
        });
    }
    get intendedReceivedAmounts() {
        return this.receivedAmounts;
    }
    get type() {
        if (this.data.vin.length === 0 || this.data.vin[0].coinbase) {
            return "coinbase";
        }
        let hasUndefined = false;
        let hasDefined = false;
        if (!this.additionalData || !this.additionalData.vinouts) {
            return "partial_payment";
        }
        for (let i = 0; i < this.additionalData.vinouts.length; i++) {
            const vinOut = this.additionalData.vinouts[i];
            if (vinOut === undefined) {
                hasUndefined = true;
            }
            else {
                hasDefined = true;
            }
        }
        if (hasDefined && !hasUndefined) {
            return "full_payment";
        }
        return "partial_payment";
    }
    get isNativePayment() {
        // On these chains there are no other types of transactions
        return true;
    }
    get elementaryUnits() {
        return (0, utils_1.toBN)(constants_1.BTC_MDU);
    }
    get successStatus() {
        return genericMccTypes_1.TransactionSuccessStatus.SUCCESS;
    }
    ///////////////////////////////
    //// Utxo specific methods ////
    ///////////////////////////////
    /**
     * Asserts whether the vin index is in valid range. If not, exception is thrown.
     * @param vinIndex vin index
     */
    assertValidVinIndex(vinIndex) {
        if (vinIndex < 0 || vinIndex >= this.sourceAddresses.length) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Invalid vin index"));
        }
    }
    /**
     * Asserts whether the vout index is in valid range. If not, exception is thrown.
     * @param voutIndex vout index
     */
    assertValidVoutIndex(voutIndex) {
        if (voutIndex < 0 || voutIndex >= this.receivingAddresses.length) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Invalid vout index"));
        }
    }
    /**
     * Extract vout on vout index. If vout index is not valid or data is corrupted, exception is thrown.
     * @param voutIndex vout index
     * @returns vout on vout index
     */
    extractVoutAt(voutIndex) {
        this.assertValidVoutIndex(voutIndex);
        const toReturn = this.data.vout[voutIndex];
        if (toReturn.n !== voutIndex) {
            throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, new Error(`Vin and vout transaction miss match; requested index ${voutIndex}, found index ${toReturn.n}`));
        }
        return toReturn;
    }
}
exports.UtxoTransaction = UtxoTransaction;
//# sourceMappingURL=UtxoTransaction.js.map